<?php

namespace Drupal\commerce_gpwebpay\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Chaching\Chaching;
use Chaching\TransactionStatuses;
use Chaching\Exceptions\InvalidResponseException;
use Drupal\commerce_payment\Exception\InvalidResponseException as CommerceInvalidResponseException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;

/**
 * Provides the Off-site Redirect payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "commerce_gpwebpay",
 *   label = "GP Webpay",
 *   display_label = @Translation("Credit Card"),
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_gpwebpay\PluginForm\PaymentGPWebpayForm",
 *   },
 *   credit_card_types = {
 *     "mastercard", "maestro", "visa", "dinersclub"
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class GPWebpay extends OffsitePaymentGatewayBase {

  /**
   * Logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private $logger;

  /**
   * Constructs a new PaymentGatewayBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, LoggerInterface $logger) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('logger.channel.commerce_gpwebpay')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'merchant_id' => '',
      'private_key_path' => '',
      'passphrase' => '',
      'certificate_path' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['merchant_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant id'),
      '#description' => $this->t('E-shop id in GP'),
      '#default_value' => $this->configuration['merchant_id'],
      '#required' => TRUE,
    ];

    $form['private_key_path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private key path'),
      '#description' => $this->t('This is a path to private key for GP Webpay.'),
      '#default_value' => $this->configuration['private_key_path'],
      '#required' => TRUE,
    ];

    $form['passphrase'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private key passphrase'),
      '#description' => $this->t('This is a passphrase that is required for the private key'),
      '#default_value' => $this->configuration['passphrase'],
      '#required' => TRUE,
    ];

    $form['certificate_path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Certificate path'),
      '#description' => $this->t('A path to the public ceritifcate of GP Webpay'),
      '#default_value' => $this->configuration['certificate_path'],
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);
    // @todo Add validation of paths.
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['merchant_id'] = $values['merchant_id'];
      $this->configuration['private_key_path'] = $values['private_key_path'];
      $this->configuration['passphrase'] = $values['passphrase'];
      $this->configuration['certificate_path'] = $values['certificate_path'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $chaching = $this->getDriverInstance();

    $request_data = $request->query;

    try {
      $gp_response = $chaching->response($request_data->all());
    }
    catch (InvalidResponseException $e) {
      watchdog_exception('commerce_gpwebpay', $e);
      $this->logger->error(print_r($request_data, TRUE));
      throw new CommerceInvalidResponseException("Failed to verify GP response");
    }

    if ($gp_response->status == TransactionStatuses::FAILURE) {
      throw new PaymentGatewayException(
        sprintf(
          "Transaction failure. PRCODE: %d SRCODE: %d",
          $request_data->get('PRCODE'), $request_data->get('SRCODE')
        )
      );
    }

    // At this point the request has been validated.
    $gp_order_id = $request_data->get('MERORDERNUM');
    if ($gp_order_id != $order->id()) {
      throw new CommerceInvalidResponseException(
        sprintf("Order number for order %d does not match %d", $order->id(), $gp_order_id)
      );
    }

    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $payment_storage->create([
      'state' => 'completed',
      'amount' => $order->getBalance(),
      'payment_gateway' => $this->parentEntity->id(),
      'order_id' => $order->id(),
      'remote_id' => $request_data->get('ORDERNUMBER'),
      'remote_state' => $request_data->get('PRCODE'),
    ]);
    $payment->save();
  }

  /**
   * Gets instance of GP Webpay driver.
   *
   * @return \Chaching\Chaching
   *   GP Webpay driver instance.
   */
  public function getDriverInstance(): Chaching {
    $driver = Chaching::GPWEBPAY;
    $authorization = [
      $this->configuration['merchant_id'],
      [
        'certificate' => $this->configuration['private_key_path'],
        'passphrase' => $this->configuration['passphrase'],
        'key' => $this->configuration['certificate_path'],
        'url_handle' => 'pgw',        
      ],
    ];
    $options = [
      'sandbox' => ($this->getMode() == 'test'),
    ];
    return new Chaching($driver, $authorization, $options);
  }

}
